﻿using PalmSens.Analysis;
using PalmSens.Core.Simplified.Data;
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace PSSDKPeakBrowserExample
{
    public class PeakData : INotifyPropertyChanged
    {
        public PeakData() { }

        public PeakData(SimpleCurve curve, SimpleMeasurement measurement, string filePath, PeakSettings peakSettings)
        {
            FileName = Path.GetFileName(filePath).Replace(".pssession", "");
            FilePath = filePath;
            Measurement = measurement.Title;
            Curve = curve.Title;
            _peakSettings = peakSettings;
            _simpleCurve = curve;
            DetectPeaks();
        }

        #region Properties
        private readonly PeakSettings _peakSettings;
        private readonly SimpleCurve _simpleCurve;
        public SimpleCurve ProcessedCurve { get; private set; }
        #endregion

        #region Functions
        public void DetectPeaks()
        {
            Peaks.Clear();
            ProcessedCurve = _simpleCurve.Clone(); //Copy curve

            if (_peakSettings.Smooth > 0) //Smooth if specified
                ProcessedCurve = _simpleCurve.Smooth(_peakSettings.Smooth);

            //Detect Peaks
            ProcessedCurve.DetectPeaks(_peakSettings.MinWidth, _peakSettings.MinHeight, true, _peakSettings.Shoulder, _peakSettings.Merge);

            //Remove peaks outside of specified range
            PeakList peaksInRange = new PeakList(ProcessedCurve.Curve);
            if (_peakSettings.WindowLeft != 0.0f || _peakSettings.WindowRight != 0.0f)
            {
                foreach (Peak p in ProcessedCurve.Peaks)
                    if (p.PeakX >= _peakSettings.WindowLeft && p.PeakX <= _peakSettings.WindowRight)
                        peaksInRange.AddPeak(p);
                ProcessedCurve.Curve.Peaks = peaksInRange;
            }

            NPeaks = ProcessedCurve.Peaks.nPeaks;
            for (int i = 0; i < ProcessedCurve.Peaks.nPeaks; i++)
                Peaks.Add($"{i + 1}");
            SelectedPeakIndex = NPeaks > 0 ? 0 : -1;
            OnPropertyChanged(nameof(NPeaks));
            OnPropertyChanged(nameof(Peaks));
        }
        #endregion

        #region Binding
        public string FileName { get; private set; }
        public string FilePath { get; private set; }
        public string Measurement { get; private set; }
        public string Curve { get; private set; }
        public int NPeaks { get; private set; }
        public ObservableCollection<string> Peaks { get; private set; } = new ObservableCollection<string>();

        private float _potential;
        public float Potential
        {
            get { return _potential; }
            private set
            {
                if (_potential == value) return;
                _potential = value;
                OnPropertyChanged(nameof(Potential));
            }
        }

        private float _current;
        public float Current
        {
            get { return _current; }
            private set
            {
                if (_current == value) return;
                _current = value;
                OnPropertyChanged(nameof(Current));
            }
        }

        private float _height;
        public float Height
        {
            get { return _height; }
            private set
            {
                if (_height == value) return;
                _height = value;
                OnPropertyChanged(nameof(Height));
            }
        }

        private float _width;
        public float Width
        {
            get { return _width; }
            private set
            {
                if (_width == value) return;
                _width = value;
                OnPropertyChanged(nameof(Width));
            }
        }

        private float _area;
        public float Area
        {
            get { return _area; }
            private set
            {
                if (_area == value) return;
                _area = value;
                OnPropertyChanged(nameof(Area));
            }
        }

        private int _selectedPeakIndex;
        public int SelectedPeakIndex
        {
            get { return _selectedPeakIndex; }
            set
            {
                _selectedPeakIndex = value;
                if (_selectedPeakIndex == -1)
                {
                    Potential = 0;
                    Current = 0;
                    Height = 0;
                    Width = 0;
                    Area = 0;
                }
                else
                {
                    Potential = (float)ProcessedCurve.Peaks[_selectedPeakIndex].PeakX;
                    Current = (float)ProcessedCurve.Peaks[_selectedPeakIndex].PeakY;
                    Height = (float)ProcessedCurve.Peaks[_selectedPeakIndex].PeakValue;
                    Width = (float)ProcessedCurve.Peaks[_selectedPeakIndex].Width;
                    Area = (float)ProcessedCurve.Peaks[_selectedPeakIndex].Area;
                }
                OnPropertyChanged(nameof(SelectedPeakIndex));
            }
        }
        #endregion

        #region Interface
        public event PropertyChangedEventHandler PropertyChanged;
        private void OnPropertyChanged(string propertyName)
        {
            PropertyChanged?.Invoke(this, new PropertyChangedEventArgs(propertyName));
        }
        #endregion
    }
}
